# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List, Optional
import logging
from operator import add
from . import EnergyDomain, EnergyDevice
from .rapl_device import RaplDevice
try :
    from .nvidia_device import NvidiaGPUDevice
except ImportError: 
    NvidiaGPUDevice=None
    logging.warning(f'pynvml not found you can\'t use NVIDIA devices') 

from ..exception import NoSuchEnergyDeviceError

from functools import reduce


class EnergyDeviceFactory:

    @staticmethod
    def _gen_all_available_domains() -> List[EnergyDevice]:
        available_api = [RaplDevice]
        if NvidiaGPUDevice is not None : 
            available_api.append(NvidiaGPUDevice)
        available_domains = []
        for api in available_api:
            try:
                available_domains.append(api.available_domains())
            except NoSuchEnergyDeviceError:
                pass
        flaten_available_domain_list = reduce(add, available_domains, [])
        return flaten_available_domain_list

    @staticmethod
    def create_devices(domains: Optional[EnergyDomain] = None) -> List[EnergyDevice]:
        """
        Create and configure the EnergyDevice instance with the given EnergyDomains
        :param domains: a list of EnergyDomain instance that as to be monitored
        :return: a list of device configured with the given EnergyDomains
        :raise NoSuchEnergyDeviceError: if a domain depend on a device that doesn't exist on the current machine
        :raise NoSuchDomainError: if the given domain is not available on the device
        """
        if domains is None:
            domains = EnergyDeviceFactory._gen_all_available_domains()

        grouped_domains = {}
        for device_type, domain in map(lambda d: (d.get_device_type(), d), domains):
            if device_type in grouped_domains:
                grouped_domains[device_type].append(domain)
            else:
                grouped_domains[device_type] = [domain]

        devices = []

        for device_type in grouped_domains:
            device = device_type()
            device.configure(domains=grouped_domains[device_type])
            devices.append(device)
        return devices



# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
from typing import List

from . import EnergyDevice, EnergyDomain, NotConfiguredDeviceException
from pyJoules.exception import NoSuchEnergyDeviceError


class RaplDomain(EnergyDomain):

    def __init__(self, socket: int):
        EnergyDomain.__init__(self)
        self.socket = socket
        self._repr = self.get_domain_name() + '_' + str(socket)

    def get_device_type(self):
        return RaplDevice

    def get_domain_name(self) -> str:
        """
        :return: domain name without socket identifier
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        """
        :return: domain name with socket identifier
        """
        return self._repr

    def __eq__(self, other) -> bool:
        return isinstance(other, RaplDomain) and self.__repr__() == other.__repr__()

    def __lt__(self, other) -> bool:
        if isinstance(other, RaplDomain):
            return self.__repr__() < other.__repr__()
        raise ValueError()

    def __gt__(self, other) -> bool:
        if isinstance(other, RaplDomain):
            return self.__repr__() > other.__repr__()
        raise ValueError()


class RaplCoreDomain(RaplDomain):
    def get_domain_name(self):
        return "core"


class RaplUncoreDomain(RaplDomain):
    def get_domain_name(self):
        return "uncore"


class RaplDramDomain(RaplDomain):
    def get_domain_name(self):
        return "dram"


class RaplPackageDomain(RaplDomain):
    def get_domain_name(self):
        return "package"


RAPL_API_DIR = '/sys/class/powercap/intel-rapl'


class RaplDevice(EnergyDevice):
    """
    Interface to get energy consumption of CPU domains
    """

    def __init__(self):
        """
        :raise NoSuchEnergyDeviceError: if no RAPL API is available on this machine
        """
        EnergyDevice.__init__(self)
        self._api_files = None

    @staticmethod
    def _rapl_api_available():
        return os.path.exists(RAPL_API_DIR)

    @staticmethod
    def available_domains() -> List[RaplDomain]:
        """
        return a the list of the available energy domains
        """
        if not RaplDevice._rapl_api_available():
            raise NoSuchEnergyDeviceError()

        return (RaplDevice.available_package_domains() + RaplDevice.available_dram_domains() +
                RaplDevice.available_core_domains() + RaplDevice.available_uncore_domains())

    @staticmethod
    def _get_socket_id_list():
        socket_id_list = []
        socket_id = 0
        while True:
            name = RAPL_API_DIR + '/intel-rapl:' + str(socket_id)
            if os.path.exists(name):
                socket_id_list.append(socket_id)
                socket_id += 1
            else:
                return socket_id_list

    @staticmethod
    def available_package_domains() -> List[RaplPackageDomain]:
        """
        return a the list of the available energy Package domains
        """
        package_domains = []

        for socket_id in RaplDevice._get_socket_id_list():
            domain_name_file_str = RAPL_API_DIR + '/intel-rapl:' + str(socket_id) + '/name'
            if os.path.exists(domain_name_file_str):
                domain_name_file = open(domain_name_file_str)
                if domain_name_file.readline() == 'package-' + str(socket_id) + '\n':
                    package_domains.append(RaplPackageDomain(socket_id))
        return package_domains

    @staticmethod
    def _domain_exist_on_socket(socket_id, domain_name):
        domain_id = 0
        while True:
            domain_name_file_str = (RAPL_API_DIR + '/intel-rapl:' + str(socket_id) + '/intel-rapl:' + str(socket_id) +
                                    ':' + str(domain_id) + '/name')
            if os.path.exists(domain_name_file_str):
                domain_name_file = open(domain_name_file_str)
                if domain_name_file.readline() == domain_name + '\n':
                    return True
                domain_id += 1
            else:
                return False

    @staticmethod
    def available_dram_domains() -> List[RaplDramDomain]:
        """
        return a the list of the available energy Dram domains
        """
        dram_domains = []
        for socket_id in RaplDevice._get_socket_id_list():
            if RaplDevice._domain_exist_on_socket(socket_id, 'dram'):
                dram_domains.append(RaplDramDomain(socket_id))
        return dram_domains

    @staticmethod
    def available_core_domains() -> List[RaplCoreDomain]:
        """
        return a the list of the available energy Core domains
        """
        core_domains = []
        for socket_id in RaplDevice._get_socket_id_list():
            if RaplDevice._domain_exist_on_socket(socket_id, 'core'):
                core_domains.append(RaplCoreDomain(socket_id))
        return core_domains

    @staticmethod
    def available_uncore_domains() -> List[RaplUncoreDomain]:
        """
        return a the list of the available energy Uncore domains
        """
        uncore_domains = []
        for socket_id in RaplDevice._get_socket_id_list():
            if RaplDevice._domain_exist_on_socket(socket_id, 'uncore'):
                uncore_domains.append(RaplUncoreDomain(socket_id))
        return uncore_domains

    def _get_domain_file_name(self, domain):
        socket_id = domain.socket

        if isinstance(domain, RaplPackageDomain):
            return RAPL_API_DIR + '/intel-rapl:' + str(socket_id) + '/energy_uj'

        domain_id = 0
        while True:
            domain_name_file_str = (RAPL_API_DIR + '/intel-rapl:' + str(socket_id) + '/intel-rapl:' + str(socket_id) +
                                    ':' + str(domain_id) + '/name')
            if os.path.exists(domain_name_file_str):
                domain_name_file = open(domain_name_file_str)
                if domain_name_file.readline() == domain.get_domain_name() + '\n':
                    return (RAPL_API_DIR + '/intel-rapl:' + str(socket_id) + '/intel-rapl:' + str(socket_id) +
                            ':' + str(domain_id) + '/energy_uj')
                else:
                    domain_id += 1
            else:
                raise ValueError()

    def _open_api_files(self, domain_list):
        api_files = []
        for domain in domain_list:
            domain_file_name = self._get_domain_file_name(domain)
            api_files.append(open(domain_file_name))
        return api_files

    def configure(self, domains=None):
        EnergyDevice.configure(self, domains)

        self._api_files = self._open_api_files(self._configured_domains)

    def _read_energy_value(self, api_file):
        api_file.seek(0, 0)
        return float(api_file.readline())

    def get_energy(self):
        return [self._read_energy_value(api_file) for api_file in self._api_files]



# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from .... utils.rapl_fs import *

from pyJoules.energy_device import NotConfiguredDeviceException
from pyJoules.energy_device.rapl_device import RaplDevice, RaplPackageDomain, RaplDramDomain, RaplCoreDomain
from pyJoules.energy_device.rapl_device import RaplUncoreDomain
from pyJoules.exception import NoSuchEnergyDeviceError, NoSuchDomainError

##############
# INIT TESTS #
##############
def test_create_RaplDevice_with_no_rapl_api_raise_NoSuchEnergyDeviceError(empty_fs):
    with pytest.raises(NoSuchEnergyDeviceError):
        RaplDevice()

def test_create_RaplDevice(fs_pkg_one_socket):
    device = RaplDevice()
    assert device is not None
    assert isinstance(device, RaplDevice)


###########################
# AVAILABLE DOMAINS TESTS #
###########################
def test_available_domains_with_no_rapl_api_raise_NoSuchEnergyDeviceError(empty_fs):
    with pytest.raises(NoSuchEnergyDeviceError):
        RaplDevice.available_domains()


def test_available_domains_with_pkg_rapl_api_return_correct_values(fs_pkg_one_socket):
    returned_values = RaplDevice.available_domains()
    correct_values = [RaplPackageDomain(0)]
    assert sorted(correct_values) == sorted(returned_values)


def test_available_domains_with_pkg_and_dram_rapl_api_return_correct_values(fs_pkg_dram_one_socket):
    returned_values = RaplDevice.available_domains()
    correct_values = [RaplPackageDomain(0), RaplDramDomain(0)]
    assert sorted(correct_values) == sorted(returned_values)

def test_available_domains_with_pkg_core_and_dram_rapl_api_return_correct_values(fs_pkg_dram_core_one_socket):
    returned_values = RaplDevice.available_domains()
    correct_values = [RaplPackageDomain(0), RaplDramDomain(0), RaplCoreDomain(0)]
    assert sorted(correct_values) == sorted(returned_values)


def test_available_domains_with_pkg_uncore_and_dram_rapl_api_return_correct_values(fs_pkg_dram_uncore_one_socket):
    returned_values = RaplDevice.available_domains()
    correct_values = [RaplPackageDomain(0), RaplDramDomain(0), RaplUncoreDomain(0)]
    assert sorted(correct_values) == sorted(returned_values)


def test_available_domains_with_pkg_psys_rapl_api_return_correct_values(fs_pkg_psys_one_socket):
    returned_values = RaplDevice.available_domains()
    correct_values = [RaplPackageDomain(0)]
    assert sorted(correct_values) == sorted(returned_values)


def test_available_domains_with_pkg_dram_rapl_api_two_cpu_return_correct_values(fs_pkg_dram_two_socket):
    returned_values = RaplDevice.available_domains()
    correct_values = [RaplPackageDomain(0), RaplDramDomain(0), RaplPackageDomain(1), RaplDramDomain(1)]
    assert sorted(correct_values) == sorted(returned_values)


###################
# CONFIGURE TESTS #
###################
def test_configure_device_to_get_dram_energy_with_no_rapl_dram_api_raise_NoSuchDomainError(fs_pkg_one_socket):
    device = RaplDevice()

    with pytest.raises(NoSuchDomainError):
        device.configure([RaplDramDomain(0)])

def test_configure_device_to_get_pkg_energy_on_cpu1_with_no_cpu1_raise_NoSuchDomainError(fs_pkg_dram_one_socket):
    device = RaplDevice()

    with pytest.raises(NoSuchDomainError):
        device.configure([RaplPackageDomain(1)])

def test_get_configured_domains_on_non_configured_device_raise_NotConfiguredDeviceException(fs_pkg_dram_one_socket):
    device = RaplDevice()

    with pytest.raises(NotConfiguredDeviceException):
        device.get_configured_domains()


def test_get_configured_domains_with_default_values_on_pkg_rapl_api_return_correct_values(fs_pkg_one_socket):
    configured_domains = [RaplPackageDomain(0)]
    device = RaplDevice()
    device.configure()
    assert configured_domains == device.get_configured_domains()


def test_get_configured_domains_with_default_values_on_pkg_dram_rapl_api_return_correct_values(fs_pkg_dram_one_socket):
    configured_domains = [RaplPackageDomain(0), RaplDramDomain(0)]
    device = RaplDevice()
    device.configure()
    assert configured_domains == device.get_configured_domains()


def test_get_configured_domains_with_pkg_values_on_pkg_dram_rapl_api_return_correct_values(fs_pkg_dram_one_socket):
    configured_domains = [RaplPackageDomain(0)]
    device = RaplDevice()
    device.configure(configured_domains)
    assert configured_domains == device.get_configured_domains()


########################################
# GET ENERGY WITH DEFAULT VALUES TESTS #
########################################
def test_get_default_energy_values_with_pkg_rapl_api(fs_pkg_one_socket):
    device = RaplDevice()
    device.configure()
    assert device.get_energy() == [fs_pkg_one_socket.domains_current_energy['package_0']]


def test_get_default_energy_values_with_pkg_dram_rapl_api(fs_pkg_dram_one_socket):
    device = RaplDevice()
    device.configure()
    assert device.get_energy() == [fs_pkg_dram_one_socket.domains_current_energy['package_0'],
                                   fs_pkg_dram_one_socket.domains_current_energy['dram_0']]


def test_get_default_energy_values_with_pkg_dram_rapl_api_two_sockets(fs_pkg_dram_two_socket):
    device = RaplDevice()
    device.configure()
    assert device.get_energy() == [fs_pkg_dram_two_socket.domains_current_energy['package_0'],
                                   fs_pkg_dram_two_socket.domains_current_energy['package_1'],
                                   fs_pkg_dram_two_socket.domains_current_energy['dram_0'],
                                   fs_pkg_dram_two_socket.domains_current_energy['dram_1']]


##################################
# CONFIGURE AND GET ENERGY TESTS #
##################################
def test_get_package_energy_with_only_pkg_rapl_api_return_correct_value(fs_pkg_one_socket):
    """
    Create a RaplDevice instance on a machine with package rapl api with on one socket
    configure it to monitor package domain
    use the `get_energy` method and check if:
    - the returned list contains one element
    - this element is the power consumption of the package on socket 0
    """
    device = RaplDevice()
    device.configure([RaplPackageDomain(0)])
    assert device.get_energy() == [fs_pkg_one_socket.domains_current_energy['package_0']]


def test_get_dram_energy_with_pkg_dram_rapl_api_return_correct_value(fs_pkg_dram_one_socket):
    """
    Create a RaplDevice instance on a machine with package and dram rapl api with on one socket
    configure it to monitor dram domain
    use the `get_energy` method and check if:
    - the returned list contains one element
    - this element is the power consumption of the dram on socket 0
    """
    device = RaplDevice()
    device.configure([RaplDramDomain(0)])
    assert device.get_energy() == [fs_pkg_dram_one_socket.domains_current_energy['dram_0']]


def test_get_package_dram_energy_with_pkg_dram_rapl_api_return_correct_value(fs_pkg_dram_one_socket):
    """
    Create a RaplDevice instance on a machine with package and dram rapl api with on one socket
    configure it to monitor package and dram domains
    use the `get_energy` method and check if:
    - the returned list contains two elements
    - these elements are the power consumption of the package and dram on socket 0
    """
    device = RaplDevice()
    device.configure([RaplPackageDomain(0), RaplDramDomain(0)])
    assert device.get_energy() == [fs_pkg_dram_one_socket.domains_current_energy['package_0'],
                                   fs_pkg_dram_one_socket.domains_current_energy['dram_0']]


def test_get_dram_package_energy_with_pkg_dram_rapl_api_return_correct_values_in_correct_order(fs_pkg_dram_one_socket):
    """
    Create a RaplDevice instance on a machine with package and dram rapl api with on one socket
    configure it to monitor dram and package domains
    use the `get_energy` method and check if:
    - the returned list contains two elements
    - these elements are the power consumption of the dram and package on socket 0
    """
    device = RaplDevice()
    device.configure([RaplDramDomain(0), RaplPackageDomain(0)])
    assert device.get_energy() == [fs_pkg_dram_one_socket.domains_current_energy['dram_0'],
                                   fs_pkg_dram_one_socket.domains_current_energy['package_0']]


def test_get_package_dram_energy_with_pkg_dram_rapl_api_two_sockets_return_correct_value(fs_pkg_dram_two_socket):
    """
    Create a RaplDevice instance on a machine with package and dram rapl api with on two socket
    configure it to monitor package and dram domains
    use the `get_energy` method and check if:
    - the returned list contains 4 elements
    - these elements are the power consumption of the package and dram on socket 0 and socket 1
    """
    device = RaplDevice()
    device.configure([RaplPackageDomain(0), RaplDramDomain(0), RaplPackageDomain(1), RaplDramDomain(1)])
    assert device.get_energy() == [fs_pkg_dram_two_socket.domains_current_energy['package_0'],
                                   fs_pkg_dram_two_socket.domains_current_energy['dram_0'],
                                   fs_pkg_dram_two_socket.domains_current_energy['package_1'],
                                   fs_pkg_dram_two_socket.domains_current_energy['dram_1']]


def test_get_package0_and_all_dram_energy_with_pkg_dram_rapl_api_two_sockets_return_correct_value(fs_pkg_dram_two_socket):
    """
    Create a RaplDevice instance on a machine with package and dram rapl api with on two socket
    configure it to monitor package domains on socket 0 and dram domains on socket 0 and 1
    use the `get_energy` method and check if:
    - the returned list contains 3 elements
    - these elements are the power consumption of the package on socket 0 and dram on socket 0 and socket 1
    """
    device = RaplDevice()
    device.configure([RaplPackageDomain(0), RaplDramDomain(0), RaplDramDomain(1)])
    assert device.get_energy() == [fs_pkg_dram_two_socket.domains_current_energy['package_0'],
                                   fs_pkg_dram_two_socket.domains_current_energy['dram_0'],
                                   fs_pkg_dram_two_socket.domains_current_energy['dram_1']]



# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import time
import operator
import functools

from functools import reduce
from typing import List, Optional

from .exception import PyJoulesException
from .energy_device import EnergyDevice, EnergyDomain, EnergyDeviceFactory
from .energy_handler import EnergyHandler, PrintHandler
from . import EnergySample

class NoNextStateException(PyJoulesException):
    """Exception raised when trying to compute duration or energy from a state
    which is the last state of an energy trace.
    """


class StateIsNotFinalError(PyJoulesException):
    """Exception raised when trying to add a state to a non final state on an energy trace
    """


class EnergyMeterNotStartedError(PyJoulesException):
    """
    Exception raised when trying to stop or record on a non started EnergyMeter instance
    """


class EnergyMeterNotStoppedError(PyJoulesException):
    """
    Exception raised when trying to get energy samples from non stopped EnergyMeter instance
    """


class SampleNotFoundError(PyJoulesException):
    """
    Exception raised when trying to retrieve a sample that does not exist on trace
    """


class EnergyMeter:
    """
    Tool used to record the energy consumption of given devices
    """

    def __init__(self, devices: List[EnergyDevice], default_tag: str = ''):
        """
        :param devices: list of the monitored devices
        :param default_tag: tag given if no tag were given to a measure
        """
        self.devices = devices
        self.default_tag = default_tag

        self._last_state = None
        self._first_state = None

    def _measure_new_state(self, tag):
        timestamp = time.perf_counter()
        values = [device.get_energy() for device in self.devices]

        return EnergyState(timestamp, tag if tag is not None else self.default_tag, values)

    def start(self, tag: Optional[str] = None):
        """
        Begin a new energy trace
        :param tag: sample name
        """
        new_state = self._measure_new_state(tag)
        self._first_state = new_state
        self._last_state = new_state

    def record(self, tag: Optional[str] = None):
        """
        Add a new state to the Trace
        :param tag: sample name
        :raise EnergyMeterNotStartedError: if the energy meter isn't started
        """
        if self._first_state is None:
            raise EnergyMeterNotStartedError()

        new_state = self._measure_new_state(tag)
        self._last_state.add_next_state(new_state)
        self._last_state = new_state

    def stop(self):
        """
        Set the end of the energy trace
        :raise EnergyMeterNotStartedError: if the energy meter isn't started
        """
        if self._first_state is None:
            raise EnergyMeterNotStartedError()

        new_state = self._measure_new_state('__stop__')
        self._last_state.add_next_state(new_state)
        self._last_state = new_state

    def get_sample(self, tag: str) -> EnergySample:
        """
        Retrieve the first sample in the trace with the given tag
        :param tag: tag of the sample to get
        :return: the sample with the given tag, if many sample have the same tag, the first sample created is returned
        :raise EnergyMeterNotStoppedError: if the energy meter isn't stopped
        :raise SampleNotFoundError: if the trace doesn't contains a sample with the given tag name
        """
        if self._first_state is None:
            raise EnergyMeterNotStartedError()

        if not self._last_state.tag == '__stop__':
            raise EnergyMeterNotStoppedError()

        for sample in self:
            if sample.tag == tag:
                return sample
        raise SampleNotFoundError()

    def _get_domain_list(self):
        """
        return the list of all monitored domains for each monitored energy devices
        """
        return reduce(operator.add, [device.get_configured_domains() for device in self.devices])

    def __iter__(self):
        """
        iterate on the energy sample of the last trace
        :raise EnergyMeterNotStoppedError: if the energy meter isn't stopped
        """
        if self._first_state is None:
            raise EnergyMeterNotStartedError()

        if not self._last_state.tag == '__stop__':
            raise EnergyMeterNotStoppedError()
        domains = self._get_domain_list()
        return SampleIterator(self._first_state, domains)

class SampleIterator:

    def __init__(self, first_state, domains):
        self.domains = domains
        self._current_state = first_state

    def _gen_sample(self, state):
        return EnergySample(state.timestamp, state.tag, state.compute_duration(), state.compute_energy(self.domains))

    def __next__(self):
        if self._current_state.next_state is None:
            raise StopIteration()

        sample = self._gen_sample(self._current_state)
        self._current_state = self._current_state.next_state
        return sample


class EnergyState:
    """
    Internal class that record the current energy state of the monitored device
    """

    def __init__(self, timestamp: float, tag: str, values: List[float]):
        """
        :param timstamp: timestamp of the measure
        :param tag: tag of the measure
        :param values: energy consumption measure, this is the list of measured energy consumption values for each
                       monitored device. This list contains the energy consumption since the last device reset to the
                       end of this sample
        """
        self.timestamp = timestamp
        self.tag = tag
        self.values = values
        self.next_state = None

    def is_last(self) -> bool:
        """
        indicate if the current state is the last state of the trace or not
         :return: True if the current state is the last state of the trace False otherwise
        """
        return self.next_state is None

    def compute_duration(self) -> float:
        """
        :return: compute the time elipsed between the current state and the next state
        :raise NoNextStateException: if the state is the last state of the trace
        """
        if self.next_state is None:
            raise NoNextStateException()

        return self.next_state.timestamp - self.timestamp

    def compute_energy(self, domains) -> List[float]:
        """
        :return: compute the energy consumed between the current state and the next state
        :raise NoNextStateException: if the state is the last state of the trace
        """
        if self.next_state is None:
            raise NoNextStateException()

        energy = []
        for next_state_device, current_state_device in zip(self.next_state.values, self.values):
            for next_value, current_value in zip(next_state_device, current_state_device):
                energy.append(next_value - current_value)

        values_dict = {}
        for value, key in zip(energy, domains):
            values_dict[str(key)] = value
        return values_dict

    def add_next_state(self, state: EnergyState):
        """
        :param previous: next state for the same energy trace
        :raise StateIsNotFinalError: if there are already a next state
        """
        if self.next_state is not None:
            raise StateIsNotFinalError()
        self.next_state = state


def measureit(func=None ,handler: EnergyHandler = PrintHandler(), domains: Optional[List[EnergyDomain]] = None):
    """
    Measure the energy consumption of monitored devices during the execution of the decorated function
    :param handler: handler instance that will receive the power consummation data
    :param domains: list of the monitored energy domains
    """
    def decorator_measure_energy(func):

        devices = EnergyDeviceFactory.create_devices(domains)
        energy_meter = EnergyMeter(devices)

        @functools.wraps(func)
        def wrapper_measure(*args, **kwargs):
            energy_meter.start(tag=func.__name__)
            val = func(*args, **kwargs)
            energy_meter.stop()
            for sample in energy_meter:
                handler.process(sample)
            return val
        return wrapper_measure

    if func is None:
        # to ensure the working system when you call it with parameters or without parameters
        return decorator_measure_energy
    else:
        return decorator_measure_energy(func)


class EnergyContext():

    def __init__(self, handler: EnergyHandler = PrintHandler(), domains: Optional[List[EnergyDomain]] = None, start_tag='start'):
        self.handler = handler
        self.start_tag = start_tag

        devices = EnergyDeviceFactory.create_devices(domains)
        self.energy_meter = EnergyMeter(devices)

    def __enter__(self) -> EnergyMeter:
        self.energy_meter.start(self.start_tag)
        return self.energy_meter

    def __exit__(self, type, value, traceback):
        self.energy_meter.stop()
        for sample in self.energy_meter:
            self.handler.process(sample)



# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from pyJoules.energy_meter import EnergyState, NoNextStateException, StateIsNotFinalError


TS_FIRST = 1.0
TS_SECOND = 2.0

E_FIRST_DOMAIN0 = 1.0
E_FIRST_DOMAIN1 = 7.2

E_SECOND_DOMAIN0 = 2.0
E_SECOND_DOMAIN1 = 9.4


@pytest.fixture
def alone_state():
    return EnergyState(TS_FIRST, 'first', [[E_FIRST_DOMAIN0, E_FIRST_DOMAIN1]])


@pytest.fixture
def two_states(alone_state):
    state2 = EnergyState(TS_SECOND, 'second', [[E_SECOND_DOMAIN0, E_SECOND_DOMAIN1]])
    alone_state.add_next_state(state2)
    return alone_state


###########
# IS_LAST #
###########
def test_new_state_are_last_sate_of_trace(alone_state):
    assert alone_state.is_last()


def test_add_a_state_to_new_state_make_it_not_being_last_state_of_trace(alone_state):
    assert alone_state.is_last()
    alone_state.add_next_state(EnergyState(1, 'second', [1, 1]))
    assert not alone_state.is_last()


###########
# COMPUTE #
###########
def test_compute_duration_from_last_state_raise_NoNextStateException(alone_state):
    with pytest.raises(NoNextStateException):
        alone_state.compute_duration()


def test_compute_energy_from_last_state_raise_NoNextStateException(alone_state):
    with pytest.raises(NoNextStateException):
        alone_state.compute_energy([])


def test_compute_duration_between_two_state_return_correct_values(two_states):
    print(two_states.next_state)
    assert two_states.compute_duration() == TS_SECOND - TS_FIRST


def test_compute_energy_between_two_state_return_correct_values(two_states):
    energy = two_states.compute_energy(['domain0', 'domain1'])
    assert len(energy) == 2
    assert energy['domain0'] == E_SECOND_DOMAIN0 - E_FIRST_DOMAIN0
    assert energy['domain1'] == E_SECOND_DOMAIN1 - E_FIRST_DOMAIN1


##################
# ADD_NEXT_STATE #
##################
def test_add_a_state_to_non_final_state_raise_StateIsNotFinalError(two_states):
    with pytest.raises(StateIsNotFinalError):
        two_states.add_next_state(EnergyState(1, 'third', [1, 2]))



# MIT License
# Copyright (c) 2019, INRIA
# Copyright (c) 2019, University of Lille
# All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List, Optional

from . import EnergyDomain
from ..exception import PyJoulesException, NoSuchDomainError

class NotConfiguredDeviceException(PyJoulesException):
    """
    Exception raised when a user call a device method that need the device to be configured on a non configured device
    """


class EnergyDevice:
    """
    Interface to get energy consumption information about a specific device
    """

    def __init__(self):
        self._configured_domains = None
        self._available_domains = self.available_domains()

    @staticmethod
    def available_domains() -> List[EnergyDomain]:
        """
        Returns names of the domain that could be monitored on the Device
        :return: a list of domain names
        :raise NoSuchEnergyDeviceError: if the device is not available on this machine
        """
        raise NotImplementedError()

    def configure(self, domains: List[EnergyDomain] = None):
        """
        configure the device to return only the energy consumption of the given energy domain when calling the
        :py:meth:`pyJoules.energy_device.EnergyDevice.get_energy` method
        :param domains: domains to be monitored by the device, if None, all the available domain will be monitored
        :raise NoSuchDomainError: if one given domain could not be monitored on this machine
        """
        if domains is None:
            domains = self._available_domains
        else:
            # check if given domain list are available
            for domain in domains:
                if domain not in self._available_domains:
                    raise NoSuchDomainError(domain)

        self._configured_domains = domains

    def get_energy(self) -> List[float]:
        """
        Get the energy consumption of the device since the last device reset
        :return: a list of each domain power consumption. Value order is the same than the domain order passed as
                 argument to the :py:meth:`pyJoules.energy_device.EnergyDevice.configure` method.
        """
        raise NotImplementedError()

    def get_configured_domains(self):
        """
        Get the domains that was passed as argument to the configure function
        :return: A list of EnergyDomain
        :raise NotConfiguredDeviceException: if the device was not configured
        """
        if self._configured_domains is None:
            raise NotConfiguredDeviceException

        return self._configured_domains
